@@ -13,7 +13,7 @@
 # querying of information, and an extension to mercurial named hgk.py,
 # which provides hooks for hgk to get information. hgk can be found in
 # the contrib directory, and hgk.py can be found in the hgext
-# directory.  
+# directory.
 #
 # To load the hgext.py extension, add it to your .hgrc file (you have
 # to use your global $HOME/.hgrc file, not one in a repository). You
@@ -45,8 +45,9 @@
 # Revisions context menu will now display additional entries to fire
 # vdiff on hovered and selected revisions.
 
-import sys, os
-from mercurial import hg, fancyopts, commands, ui, util, patch, revlog
+import os
+from mercurial import commands, util, patch, revlog
+from mercurial.node import nullid, nullrev, short
 
 def difftree(ui, repo, node1=None, node2=None, *files, **opts):
     """diff trees from two commits"""
@@ -57,21 +58,18 @@ def difftree(ui, repo, node1=None, node2=None, *files, **opts):
         status = repo.status(node1, node2, files=files)[:5]
         modified, added, removed, deleted, unknown = status
 
-        empty = hg.short(hg.nullid)
+        empty = short(nullid)
 
         for f in modified:
             # TODO get file permissions
-            print ":100664 100664 %s %s M\t%s\t%s" % (hg.short(mmap[f]),
-                                                      hg.short(mmap2[f]),
-                                                      f, f)
+            ui.write(":100664 100664 %s %s M\t%s\t%s\n" %
+                     (short(mmap[f]), short(mmap2[f]), f, f))
         for f in added:
-            print ":000000 100664 %s %s N\t%s\t%s" % (empty,
-                                                      hg.short(mmap2[f]),
-                                                      f, f)
+            ui.write(":000000 100664 %s %s N\t%s\t%s\n" %
+                     (empty, short(mmap2[f]), f, f))
         for f in removed:
-            print ":100664 000000 %s %s D\t%s\t%s" % (hg.short(mmap[f]),
-                                                      empty,
-                                                      f, f)
+            ui.write(":100664 000000 %s %s D\t%s\t%s\n" %
+                     (short(mmap[f]), empty, f, f))
     ##
 
     while True:
@@ -93,7 +91,7 @@ def difftree(ui, repo, node1=None, node2=None, *files, **opts):
             node1 = repo.changelog.parents(node1)[0]
         if opts['patch']:
             if opts['pretty']:
-                catcommit(repo, node2, "")
+                catcommit(ui, repo, node2, "")
             patch.diff(repo, node1, node2,
                        files=files,
                        opts=patch.diffopts(ui, {'git': True}))
@@ -102,14 +100,14 @@ def difftree(ui, repo, node1=None, node2=None, *files, **opts):
         if not opts['stdin']:
             break
 
-def catcommit(repo, n, prefix, ctx=None):
+def catcommit(ui, repo, n, prefix, ctx=None):
     nlprefix = '\n' + prefix;
     if ctx is None:
         ctx = repo.changectx(n)
     (p1, p2) = ctx.parents()
-    print "tree %s" % (hg.short(ctx.changeset()[0])) # use ctx.node() instead ??
-    if p1: print "parent %s" % (hg.short(p1.node()))
-    if p2: print "parent %s" % (hg.short(p2.node()))
+    ui.write("tree %s\n" % short(ctx.changeset()[0])) # use ctx.node() instead ??
+    if p1: ui.write("parent %s\n" % short(p1.node()))
+    if p2: ui.write("parent %s\n" % short(p2.node()))
     date = ctx.date()
     description = ctx.description().replace("\0", "")
     lines = description.splitlines()
@@ -118,23 +116,24 @@ def catcommit(repo, n, prefix, ctx=None):
     else:
         committer = ctx.user()
 
-    print "author %s %s %s" % (ctx.user(), int(date[0]), date[1])
-    print "committer %s %s %s" % (committer, int(date[0]), date[1])
-    print "revision %d" % ctx.rev()
-    print ""
+    ui.write("author %s %s %s\n" % (ctx.user(), int(date[0]), date[1]))
+    ui.write("committer %s %s %s\n" % (committer, int(date[0]), date[1]))
+    ui.write("revision %d\n" % ctx.rev())
+    ui.write("branch %s\n\n" % ctx.branch())
+
     if prefix != "":
-        print "%s%s" % (prefix, description.replace('\n', nlprefix).strip())
+        ui.write("%s%s\n" % (prefix, description.replace('\n', nlprefix).strip()))
     else:
-        print description
+        ui.write(description + "\n")
     if prefix:
-        sys.stdout.write('\0')
+        ui.write('\0')
 
 def base(ui, repo, node1, node2):
     """Output common ancestor information"""
     node1 = repo.lookup(node1)
     node2 = repo.lookup(node2)
     n = repo.changelog.ancestor(node1, node2)
-    print hg.short(n)
+    ui.write(short(n) + "\n")
 
 def catfile(ui, repo, type=None, r=None, **opts):
     """cat a specific revision"""
@@ -157,10 +156,10 @@ def catfile(ui, repo, type=None, r=None, **opts):
 
     while r:
         if type != "commit":
-            sys.stderr.write("aborting hg cat-file only understands commits\n")
-            sys.exit(1);
+            ui.warn("aborting hg cat-file only understands commits\n")
+            return 1;
         n = repo.lookup(r)
-        catcommit(repo, n, prefix)
+        catcommit(ui, repo, n, prefix)
         if opts['stdin']:
             try:
                 (type, r) = raw_input().split(' ');
@@ -174,7 +173,7 @@ def catfile(ui, repo, type=None, r=None, **opts):
 # telling you which commits are reachable from the supplied ones via
 # a bitmask based on arg position.
 # you can specify a commit to stop at by starting the sha1 with ^
-def revtree(args, repo, full="tree", maxnr=0, parents=False):
+def revtree(ui, args, repo, full="tree", maxnr=0, parents=False):
     def chlogwalk():
         count = repo.changelog.count()
         i = count
@@ -254,29 +253,29 @@ def revtree(args, repo, full="tree", maxnr=0, parents=False):
             parentstr = ""
             if parents:
                 pp = repo.changelog.parents(n)
-                if pp[0] != hg.nullid:
-                    parentstr += " " + hg.short(pp[0])
-                if pp[1] != hg.nullid:
-                    parentstr += " " + hg.short(pp[1])
+                if pp[0] != nullid:
+                    parentstr += " " + short(pp[0])
+                if pp[1] != nullid:
+                    parentstr += " " + short(pp[1])
             if not full:
-                print hg.short(n) + parentstr
+                ui.write("%s%s\n" % (short(n), parentstr))
             elif full == "commit":
-                print hg.short(n) + parentstr
-                catcommit(repo, n, '    ', ctx)
+                ui.write("%s%s\n" % (short(n), parentstr))
+                catcommit(ui, repo, n, '    ', ctx)
             else:
                 (p1, p2) = repo.changelog.parents(n)
-                (h, h1, h2) = map(hg.short, (n, p1, p2))
+                (h, h1, h2) = map(short, (n, p1, p2))
                 (i1, i2) = map(repo.changelog.rev, (p1, p2))
 
                 date = ctx.date()[0]
-                print "%s %s:%s" % (date, h, mask),
+                ui.write("%s %s:%s" % (date, h, mask))
                 mask = is_reachable(want_sha1, reachable, p1)
-                if i1 != hg.nullrev and mask > 0:
-                    print "%s:%s " % (h1, mask),
+                if i1 != nullrev and mask > 0:
+                    ui.write("%s:%s " % (h1, mask)),
                 mask = is_reachable(want_sha1, reachable, p2)
-                if i2 != hg.nullrev and mask > 0:
-                    print "%s:%s " % (h2, mask),
-                print ""
+                if i2 != nullrev and mask > 0:
+                    ui.write("%s:%s " % (h2, mask))
+                ui.write("\n")
             if maxnr and count >= maxnr:
                 break
             count += 1
@@ -304,15 +303,15 @@ def revlist(ui, repo, *revs, **opts):
     else:
         full = None
     copy = [x for x in revs]
-    revtree(copy, repo, full, opts['max_count'], opts['parents'])
+    revtree(ui, copy, repo, full, opts['max_count'], opts['parents'])
 
 def config(ui, repo, **opts):
     """print extension options"""
     def writeopt(name, value):
-        ui.write('k=%s\nv=%s\n' % (name, value)) 
+        ui.write('k=%s\nv=%s\n' % (name, value))
 
     writeopt('vdiff', ui.config('hgk', 'vdiff', ''))
-    
+
 
 def view(ui, repo, *etc, **opts):
     "start interactive history viewer"
