@@ -7,3 +7,6 @@ doctest.testmod(mercurial.changelog)
 
 import mercurial.httprepo
 doctest.testmod(mercurial.httprepo)
+
+import mercurial.util
+doctest.testmod(mercurial.util)
